var colorNotebook = {
    "playgroundWheels": ['Pygments-2.12.0-py3-none-any.whl', 'coloraide-2.6-py3-none-any.whl'],
    "notebookWheels": ['pyyaml', 'Markdown-3.3.4-py3-none-any.whl', 'pymdown_extensions-9.10-py3-none-any.whl', 'Pygments-2.12.0-py3-none-any.whl', 'coloraide-2.6-py3-none-any.whl'],
    "defaultPlayground": "import coloraide\ncoloraide.__version__\nColor('red')"
}
